@@ -38,6 +38,7 @@ static gboolean gconf_handle_corba_exception(CORBA_Environment* ev, GError** err
 /* just returns TRUE if there's an exception indicating the server is
    probably hosed; no side effects */
 static gboolean gconf_server_broken(CORBA_Environment* ev);
+static void gconf_detach_config_server(void);
 
 /* Maximum number of times to try re-spawning the server if it's down. */
 #define MAX_RETRIES 1
@@ -47,11 +48,19 @@ gconf_key_check(const gchar* key, GError** err)
 {
   gchar* why = NULL;
 
-  if (!gconf_valid_key(key, &why))
+  if (key == NULL)
     {
       if (err)
-        *err = gconf_error_new(GCONF_ERROR_BAD_KEY, _("`%s': %s"),
-                               key, why);
+        *err = gconf_error_new (GCONF_ERROR_BAD_KEY,
+                                _("Key \"%s\" is NULL"),
+                                key);
+      return FALSE;
+    }
+  else if (!gconf_valid_key (key, &why))
+    {
+      if (err)
+        *err = gconf_error_new (GCONF_ERROR_BAD_KEY, _("\"%s\": %s"),
+                                key, why);
       g_free(why);
       return FALSE;
     }
@@ -78,6 +87,9 @@ struct _GConfEngine {
 
   gpointer user_data;
   GDestroyNotify dnotify;
+
+  gpointer owner;
+  int owner_use_count;
   
   guint is_default : 1;
 
@@ -115,7 +127,6 @@ static ConfigServer   gconf_get_config_server    (gboolean     start_if_not_foun
 
 /* Forget our current server object reference, so the next call to
    gconf_get_config_server will have to try to respawn the server */
-static void           gconf_detach_config_server (void);
 static ConfigListener gconf_get_config_listener  (void);
 
 static void           gconf_engine_detach       (GConfEngine     *conf);
@@ -129,6 +140,11 @@ static ConfigDatabase gconf_engine_get_database (GConfEngine     *conf,
                                                  GError         **err);
 
 
+#define CHECK_OWNER_USE(engine)   \
+  do { if ((engine)->owner && (engine)->owner_use_count == 0) \
+     g_warning ("You can't use a GConfEngine that has an active GConfClient wrapper object. Use GConfClient API instead.");  \
+  } while (0)
+
 static void         register_engine           (GConfEngine    *conf);
 static void         unregister_engine         (GConfEngine    *conf);
 static GConfEngine *lookup_engine             (const gchar    *address);
@@ -173,6 +189,9 @@ gconf_engine_blank (gboolean remote)
 
   conf->refcount = 1;
   
+  conf->owner = NULL;
+  conf->owner_use_count = 0;
+  
   if (remote)
     {
       conf->database = CORBA_OBJECT_NIL;
@@ -189,10 +208,38 @@ gconf_engine_blank (gboolean remote)
       conf->is_local = TRUE;
       conf->is_default = FALSE;
     }
-    
+  
   return conf;
 }
 
+void
+gconf_engine_set_owner (GConfEngine *engine,
+                        gpointer     client)
+{
+  g_return_if_fail (engine->owner_use_count == 0);
+  
+  engine->owner = client;
+}
+
+void
+gconf_engine_push_owner_usage (GConfEngine *engine,
+                               gpointer     client)
+{
+  g_return_if_fail (engine->owner == client);
+
+  engine->owner_use_count += 1;
+}
+
+void
+gconf_engine_pop_owner_usage  (GConfEngine *engine,
+                               gpointer     client)
+{
+  g_return_if_fail (engine->owner == client);
+  g_return_if_fail (engine->owner_use_count > 0);
+
+  engine->owner_use_count -= 1;
+}
+
 static GHashTable *engines_by_db = NULL;
 
 static GConfEngine *
@@ -204,6 +251,20 @@ lookup_engine_by_database (ConfigDatabase db)
     return NULL;
 }
 
+static void
+database_rec_release (gpointer rec)
+{
+  GConfEngine *conf = rec;
+  CORBA_Environment ev;
+
+  CORBA_exception_init (&ev);
+
+  CORBA_Object_release (conf->database, &ev);
+  conf->database = CORBA_OBJECT_NIL;
+  
+  CORBA_exception_free (&ev);
+}
+
 /* This takes ownership of the ConfigDatabase */
 static void
 gconf_engine_set_database (GConfEngine *conf,
@@ -214,8 +275,11 @@ gconf_engine_set_database (GConfEngine *conf,
   conf->database = db;
 
   if (engines_by_db == NULL)
-    engines_by_db = g_hash_table_new ((GHashFunc) gconf_CORBA_Object_hash,
-                                      (GCompareFunc) gconf_CORBA_Object_equal);
+    engines_by_db = g_hash_table_new_full (
+	    (GHashFunc) gconf_CORBA_Object_hash,
+	    (GCompareFunc) gconf_CORBA_Object_equal,
+	    NULL,
+	    database_rec_release);
   
   g_hash_table_insert (engines_by_db, conf->database, conf);  
 }
@@ -223,16 +287,9 @@ gconf_engine_set_database (GConfEngine *conf,
 static void
 gconf_engine_detach (GConfEngine *conf)
 {
-  CORBA_Environment ev;
-
-  CORBA_exception_init (&ev);
-  
-  if (!CORBA_Object_is_nil (conf->database, &ev))
+  if (conf->database != CORBA_OBJECT_NIL)
     {
       g_hash_table_remove (engines_by_db, conf->database);
-      
-      CORBA_Object_release (conf->database, &ev);
-      conf->database = CORBA_OBJECT_NIL;
     }
 }
 
@@ -264,7 +321,7 @@ gconf_engine_connect (GConfEngine *conf,
     db = ConfigServer_get_default_database (cs, &ev);      
   else
     db = ConfigServer_get_database (cs, conf->address, &ev);
-  
+
   if (gconf_server_broken(&ev))
     {
       if (tries < MAX_RETRIES)
@@ -394,10 +451,11 @@ gconf_engine_get_default (void)
 
       default_engine = conf;
       
-      /* Ignore errors, we never return a NULL
-       * default database
+      /* Ignore errors, we never return a NULL default database, and
+       * since we aren't starting if it isn't found, we'll probably
+       * get errors half the time anyway.
        */
-      gconf_engine_connect (conf, TRUE, NULL);
+      gconf_engine_connect (conf, FALSE, NULL);
     }
   else
     conf->refcount += 1;
@@ -436,7 +494,7 @@ gconf_engine_get_for_address (const gchar* address, GError** err)
 }
 
 void
-gconf_engine_ref             (GConfEngine* conf)
+gconf_engine_ref(GConfEngine* conf)
 {
   g_return_if_fail(conf != NULL);
   g_return_if_fail(conf->refcount > 0);
@@ -445,7 +503,7 @@ gconf_engine_ref             (GConfEngine* conf)
 }
 
 void         
-gconf_engine_unref        (GConfEngine* conf)
+gconf_engine_unref(GConfEngine* conf)
 {
   g_return_if_fail(conf != NULL);
   g_return_if_fail(conf->refcount > 0);
@@ -561,8 +619,13 @@ gconf_engine_notify_add(GConfEngine* conf,
   CORBA_Environment ev;
   GConfCnxn* cnxn;
   gint tries = 0;
-
+  ConfigDatabase3_PropList properties;
+#define NUM_PROPERTIES 1
+  ConfigStringProperty properties_buffer[1];
+  
   g_return_val_if_fail(!gconf_engine_is_local(conf), 0);
+
+  CHECK_OWNER_USE (conf);
   
   if (gconf_engine_is_local(conf))
     {
@@ -572,6 +635,16 @@ gconf_engine_notify_add(GConfEngine* conf,
 
       return 0;
     }
+
+  properties._buffer = properties_buffer;
+  properties._length = NUM_PROPERTIES;
+  properties._maximum = NUM_PROPERTIES;
+  properties._release = CORBA_FALSE; /* don't free static buffer */
+
+  properties._buffer[0].key = "name";
+  properties._buffer[0].value = g_get_prgname ();
+  if (properties._buffer[0].value == NULL)
+    properties._buffer[0].value = "unknown";
   
   CORBA_exception_init(&ev);
 
@@ -586,10 +659,24 @@ gconf_engine_notify_add(GConfEngine* conf,
   
   /* Should have aborted the program in this case probably */
   g_return_val_if_fail(cl != CORBA_OBJECT_NIL, 0);
-
-  id = ConfigDatabase_add_listener(db,
-                                   (gchar*)namespace_section, 
-                                   cl, &ev);
+  
+  id = ConfigDatabase3_add_listener_with_properties (db,
+                                                     (gchar*)namespace_section, 
+                                                     cl,
+                                                     &properties,
+                                                     &ev);
+  
+  if (ev._major == CORBA_SYSTEM_EXCEPTION &&
+      CORBA_exception_id (&ev) &&
+      strcmp (CORBA_exception_id (&ev), "IDL:CORBA/BAD_OPERATION:1.0") == 0)
+    {
+      CORBA_exception_free (&ev);
+      CORBA_exception_init (&ev);      
+  
+      id = ConfigDatabase_add_listener(db,
+                                       (gchar*)namespace_section, 
+                                       cl, &ev);
+    }
   
   if (gconf_server_broken(&ev))
     {
@@ -614,13 +701,15 @@ gconf_engine_notify_add(GConfEngine* conf,
 
 void         
 gconf_engine_notify_remove(GConfEngine* conf,
-                    guint client_id)
+                           guint client_id)
 {
   GConfCnxn* gcnxn;
   CORBA_Environment ev;
   ConfigDatabase db;
   gint tries = 0;
 
+  CHECK_OWNER_USE (conf);
+  
   if (gconf_engine_is_local(conf))
     return;
   
@@ -688,6 +777,8 @@ gconf_engine_get_fuller (GConfEngine *conf,
   g_return_val_if_fail(key != NULL, NULL);
   g_return_val_if_fail(err == NULL || *err == NULL, NULL);
 
+  CHECK_OWNER_USE (conf);
+  
   if (!gconf_key_check(key, err))
     return NULL;
 
@@ -845,6 +936,8 @@ gconf_engine_get_entry(GConfEngine* conf,
   GConfEntry *entry;
   gchar *schema_name;
 
+  CHECK_OWNER_USE (conf);
+  
   schema_name = NULL;
   error = NULL;
   val = gconf_engine_get_fuller (conf, key, locale, use_schema_default,
@@ -902,6 +995,8 @@ gconf_engine_get_default_from_schema (GConfEngine* conf,
   g_return_val_if_fail(conf != NULL, NULL);
   g_return_val_if_fail(key != NULL, NULL);
   g_return_val_if_fail(err == NULL || *err == NULL, NULL);
+
+  CHECK_OWNER_USE (conf);
   
   if (!gconf_key_check(key, err))
     return NULL;
@@ -987,6 +1082,8 @@ gconf_engine_set (GConfEngine* conf, const gchar* key,
   g_return_val_if_fail( (value->type != GCONF_VALUE_LIST) ||
                         (gconf_value_get_list_type(value) != GCONF_VALUE_INVALID), FALSE);
   g_return_val_if_fail(err == NULL || *err == NULL, FALSE);
+
+  CHECK_OWNER_USE (conf);
   
   if (!gconf_key_check(key, err))
     return FALSE;
@@ -1057,15 +1154,17 @@ gconf_engine_set (GConfEngine* conf, const gchar* key,
 }
 
 gboolean
-gconf_engine_unset(GConfEngine* conf, const gchar* key, GError** err)
+gconf_engine_unset (GConfEngine* conf, const gchar* key, GError** err)
 {
   CORBA_Environment ev;
   ConfigDatabase db;
   gint tries = 0;
 
-  g_return_val_if_fail(conf != NULL, FALSE);
-  g_return_val_if_fail(key != NULL, FALSE);
-  g_return_val_if_fail(err == NULL || *err == NULL, FALSE);
+  g_return_val_if_fail (conf != NULL, FALSE);
+  g_return_val_if_fail (key != NULL, FALSE);
+  g_return_val_if_fail (err == NULL || *err == NULL, FALSE);
+
+  CHECK_OWNER_USE (conf);
   
   if (!gconf_key_check(key, err))
     return FALSE;
@@ -1105,11 +1204,11 @@ gconf_engine_unset(GConfEngine* conf, const gchar* key, GError** err)
       return FALSE;
     }
 
-  ConfigDatabase_unset(db,
-                     (gchar*)key,
-                     &ev);
+  ConfigDatabase_unset (db,
+                        (gchar*)key,
+                        &ev);
 
-  if (gconf_server_broken(&ev))
+  if (gconf_server_broken (&ev))
     {
       if (tries < MAX_RETRIES)
         {
@@ -1120,10 +1219,104 @@ gconf_engine_unset(GConfEngine* conf, const gchar* key, GError** err)
         }
     }
   
-  if (gconf_handle_corba_exception(&ev, err))
+  if (gconf_handle_corba_exception (&ev, err))
     return FALSE;
 
-  g_return_val_if_fail(err == NULL || *err == NULL, FALSE);
+  g_return_val_if_fail (err == NULL || *err == NULL, FALSE);
+  
+  return TRUE;
+}
+
+/**
+ * gconf_engine_recursive_unset:
+ * @engine: a #GConfEngine
+ * @key: a key or directory name
+ * @flags: change how the unset is done
+ * @err: return location for a #GError, or %NULL to ignore errors
+ * 
+ * Unsets all keys below @key, including @key itself.  If any unset
+ * fails, continues on to unset as much as it can. The first
+ * failure is returned in @err.
+ *
+ * Returns: %FALSE if error is set
+ **/
+gboolean
+gconf_engine_recursive_unset (GConfEngine    *conf,
+                              const char     *key,
+                              GConfUnsetFlags flags,
+                              GError        **err)
+{
+  CORBA_Environment ev;
+  ConfigDatabase3 db;
+  gint tries = 0;
+  ConfigDatabase3_UnsetFlags corba_flags;
+  
+  g_return_val_if_fail (conf != NULL, FALSE);
+  g_return_val_if_fail (key != NULL, FALSE);
+  g_return_val_if_fail (err == NULL || *err == NULL, FALSE);
+
+  CHECK_OWNER_USE (conf);
+  
+  if (!gconf_key_check (key, err))
+    return FALSE;
+
+  if (gconf_engine_is_local (conf))
+    {
+      GError* error = NULL;
+      
+      gconf_sources_recursive_unset (conf->local_sources, key, NULL,
+                                     flags, NULL, &error);
+
+      if (error != NULL)
+        {
+          if (err)
+            *err = error;
+          else
+            {
+              g_error_free (error);
+            }
+          return FALSE;
+        }
+      
+      return TRUE;
+    }
+
+  g_assert (!gconf_engine_is_local (conf));
+  
+  CORBA_exception_init(&ev);
+
+  corba_flags = 0;
+  if (flags & GCONF_UNSET_INCLUDING_SCHEMA_NAMES)
+    corba_flags |= ConfigDatabase3_UNSET_INCLUDING_SCHEMA_NAMES;
+  
+ RETRY:
+  
+  db = (ConfigDatabase3) gconf_engine_get_database (conf, TRUE, err);
+
+  if (db == CORBA_OBJECT_NIL)
+    {
+      g_return_val_if_fail (err == NULL || *err != NULL, FALSE);
+
+      return FALSE;
+    }
+
+  ConfigDatabase3_recursive_unset (db, key, corba_flags, &ev);
+
+  if (gconf_server_broken (&ev))
+    {
+      if (tries < MAX_RETRIES)
+        {
+          ++tries;
+          CORBA_exception_free(&ev);
+          gconf_engine_detach(conf);
+          goto RETRY;
+        }
+    }
+  
+  if (gconf_handle_corba_exception (&ev, err))
+    return FALSE;
+
+  g_return_val_if_fail (err == NULL || *err == NULL, FALSE);
   
   return TRUE;
 }
@@ -1136,22 +1329,23 @@ gconf_engine_associate_schema  (GConfEngine* conf, const gchar* key,
   ConfigDatabase db;
   gint tries = 0;
 
-  g_return_val_if_fail(conf != NULL, FALSE);
-  g_return_val_if_fail(key != NULL, FALSE);
-  g_return_val_if_fail(schema_key != NULL, FALSE);
-  g_return_val_if_fail(err == NULL || *err == NULL, FALSE);
+  g_return_val_if_fail (conf != NULL, FALSE);
+  g_return_val_if_fail (key != NULL, FALSE);
+  g_return_val_if_fail (err == NULL || *err == NULL, FALSE);
+
+  CHECK_OWNER_USE (conf);
   
-  if (!gconf_key_check(key, err))
+  if (!gconf_key_check (key, err))
     return FALSE;
 
-  if (!gconf_key_check(schema_key, err))
+  if (schema_key && !gconf_key_check (schema_key, err))
     return FALSE;
 
   if (gconf_engine_is_local(conf))
     {
       GError* error = NULL;
       
-      gconf_sources_set_schema(conf->local_sources, key, schema_key, &error);
+      gconf_sources_set_schema (conf->local_sources, key, schema_key, &error);
 
       if (error != NULL)
         {
@@ -1167,32 +1361,33 @@ gconf_engine_associate_schema  (GConfEngine* conf, const gchar* key,
       return TRUE;
     }
 
-  g_assert(!gconf_engine_is_local(conf));
+  g_assert (!gconf_engine_is_local (conf));
   
-  CORBA_exception_init(&ev);
+  CORBA_exception_init (&ev);
 
  RETRY:
   
-  db = gconf_engine_get_database(conf, TRUE, err);
+  db = gconf_engine_get_database (conf, TRUE, err);
 
   if (db == CORBA_OBJECT_NIL)
     {
-      g_return_val_if_fail(err == NULL || *err != NULL, FALSE);
+      g_return_val_if_fail (err == NULL || *err != NULL, FALSE);
 
       return FALSE;
     }
 
-  ConfigDatabase_set_schema(db,
-                            (gchar*)key,
-                            (gchar*)schema_key,
-                            &ev);
+  ConfigDatabase_set_schema (db,
+                             key,
+                             /* empty string means unset */
+                             schema_key ? schema_key : "",
+                             &ev);
 
-  if (gconf_server_broken(&ev))
+  if (gconf_server_broken (&ev))
     {
       if (tries < MAX_RETRIES)
         {
           ++tries;
-          CORBA_exception_free(&ev);
+          CORBA_exception_free (&ev);
           gconf_engine_detach (conf);
           goto RETRY;
         }
@@ -1201,7 +1396,7 @@ gconf_engine_associate_schema  (GConfEngine* conf, const gchar* key,
   if (gconf_handle_corba_exception(&ev, err))
     return FALSE;
 
-  g_return_val_if_fail(*err == NULL, FALSE);
+  g_return_val_if_fail (*err == NULL, FALSE);
   
   return TRUE;
 }
@@ -1242,6 +1437,8 @@ gconf_engine_all_entries(GConfEngine* conf, const gchar* dir, GError** err)
   g_return_val_if_fail(conf != NULL, NULL);
   g_return_val_if_fail(dir != NULL, NULL);
   g_return_val_if_fail(err == NULL || *err == NULL, NULL);
+
+  CHECK_OWNER_USE (conf);
   
   if (!gconf_key_check(dir, err))
     return NULL;
@@ -1389,6 +1586,8 @@ gconf_engine_all_dirs(GConfEngine* conf, const gchar* dir, GError** err)
   g_return_val_if_fail(conf != NULL, NULL);
   g_return_val_if_fail(dir != NULL, NULL);
   g_return_val_if_fail(err == NULL || *err == NULL, NULL);
+
+  CHECK_OWNER_USE (conf);
   
   if (!gconf_key_check(dir, err))
     return NULL;
@@ -1483,6 +1682,8 @@ gconf_engine_suggest_sync(GConfEngine* conf, GError** err)
   g_return_if_fail(conf != NULL);
   g_return_if_fail(err == NULL || *err == NULL);
 
+  CHECK_OWNER_USE (conf);
+  
   if (gconf_engine_is_local(conf))
     {
       GError* error = NULL;
@@ -1546,6 +1747,11 @@ gconf_clear_cache(GConfEngine* conf, GError** err)
   g_return_if_fail(conf != NULL);
   g_return_if_fail(err == NULL || *err == NULL);
 
+  /* don't disallow non-owner use here since you can't do this
+   * via GConfClient API and calling this function won't break
+   * GConfClient anyway
+   */
+  
   if (gconf_engine_is_local(conf))
     {
       GError* error = NULL;
@@ -1671,6 +1877,8 @@ gconf_engine_dir_exists(GConfEngine *conf, const gchar *dir, GError** err)
   g_return_val_if_fail(conf != NULL, FALSE);
   g_return_val_if_fail(dir != NULL, FALSE);
   g_return_val_if_fail(err == NULL || *err == NULL, FALSE);
+
+  CHECK_OWNER_USE (conf);
   
   if (!gconf_key_check(dir, err))
     return FALSE;
@@ -1730,6 +1938,9 @@ gconf_engine_remove_dir (GConfEngine* conf,
   g_return_if_fail(conf != NULL);
   g_return_if_fail(dir != NULL);
   g_return_if_fail(err == NULL || *err == NULL);
+
+  /* FIXME we have no GConfClient method for doing this */
+  /*   CHECK_OWNER_USE (conf); */
   
   if (!gconf_key_check(dir, err))
     return;
@@ -1777,6 +1988,8 @@ gconf_engine_key_is_writable  (GConfEngine *conf,
   gboolean is_writable = TRUE;
   GConfValue *val;
 
+  CHECK_OWNER_USE (conf);
+  
   /* FIXME implement IDL to allow getting only writability
    * (not that urgent since GConfClient caches this crap
    * anyway)
@@ -1862,13 +2075,14 @@ try_to_contact_server (gboolean start_if_not_found,
       
       if (ev._major != CORBA_NO_EXCEPTION)
 	{
-	  server = CORBA_OBJECT_NIL;
           g_set_error (err,
                        GCONF_ERROR,
                        GCONF_ERROR_NO_SERVER,
                        _("Adding client to server's list failed, CORBA error: %s"),
                        CORBA_exception_id (&ev));
 
+	  CORBA_Object_release (server, &ev);
+	  server = CORBA_OBJECT_NIL;
           CORBA_exception_free(&ev);
 	}
     }
@@ -1896,29 +2110,58 @@ gconf_get_config_server(gboolean start_if_not_found, GError** err)
   return server; /* return what we have, NIL or not */
 }
 
-static void
+ConfigListener listener = CORBA_OBJECT_NIL;
+
+void
 gconf_detach_config_server(void)
 {  
+  CORBA_Environment ev;
+
+  CORBA_exception_init(&ev);
+
+  if (listener != CORBA_OBJECT_NIL)
+    {
+      CORBA_Object_release(listener, &ev);
+      listener = CORBA_OBJECT_NIL;
+    }
+
   if (server != CORBA_OBJECT_NIL)
     {
-      CORBA_Environment ev;
-      
-      CORBA_exception_init(&ev);
-
       CORBA_Object_release(server, &ev);
 
       if (ev._major != CORBA_NO_EXCEPTION)
         {
           g_warning("Exception releasing gconfd server object: %s",
                     CORBA_exception_id(&ev));
-          CORBA_exception_free(&ev);
         }
 
       server = CORBA_OBJECT_NIL;
     }
+
+  CORBA_exception_free(&ev);
+
+  if (engines_by_db != NULL)
+    {
+      g_hash_table_destroy (engines_by_db);
+      engines_by_db = NULL;
+    }
+}
+
+/**
+ * gconf_debug_shutdown:
+ * @void: 
+ * 
+ * Detach from the config server and release
+ * all related resources
+ **/
+int
+gconf_debug_shutdown (void)
+{
+  gconf_detach_config_server ();
+
+  return gconf_orb_release ();
 }
 
-ConfigListener listener = CORBA_OBJECT_NIL;
 static void notify                  (PortableServer_Servant     servant,
                                      ConfigDatabase             db,
                                      CORBA_unsigned_long        cnxn,
@@ -2096,14 +2339,12 @@ drop_all_caches (PortableServer_Servant     _servant,
 
 static ConfigListener 
 gconf_get_config_listener(void)
-{
-  static ConfigListener listener = CORBA_OBJECT_NIL;
-  
+{  
   if (listener == CORBA_OBJECT_NIL)
     {
       CORBA_Environment ev;
-      PortableServer_ObjectId* objid;
       PortableServer_POA poa;
+      PortableServer_POAManager poa_mgr;
 
       CORBA_exception_init (&ev);
       POA_ConfigListener__init (&poa_listener_servant, &ev);
@@ -2116,18 +2357,18 @@ gconf_get_config_listener(void)
 
       g_assert (ev._major == CORBA_NO_EXCEPTION);
 
-      PortableServer_POAManager_activate(PortableServer_POA__get_the_POAManager(poa, &ev), &ev);
+      poa_mgr = PortableServer_POA__get_the_POAManager (poa, &ev);
+      PortableServer_POAManager_activate (poa_mgr, &ev);
 
       g_assert (ev._major == CORBA_NO_EXCEPTION);
 
-      objid = PortableServer_POA_activate_object (poa, &poa_listener_servant, &ev);
-
-      g_assert (ev._major == CORBA_NO_EXCEPTION);
-      
       listener = PortableServer_POA_servant_to_reference(poa,
                                                          &poa_listener_servant,
                                                          &ev);
 
+      CORBA_Object_release ((CORBA_Object) poa_mgr, &ev);
+      CORBA_Object_release ((CORBA_Object) poa, &ev);
+
       g_assert (listener != CORBA_OBJECT_NIL);
       g_assert (ev._major == CORBA_NO_EXCEPTION);
     }
@@ -2385,8 +2626,13 @@ gconf_unescape_key (const char *escaped_key,
 gboolean
 gconf_key_is_below   (const gchar* above, const gchar* below)
 {
-  int len = strlen(above);
-  if (strncmp(below, above, len) == 0)
+  int len;
+
+  if (above[0] == '/' && above[1] == '\0')
+    return TRUE;
+  
+  len = strlen (above);
+  if (strncmp (below, above, len) == 0)
     {
       /* only if this is a complete key component,
        * so that /foo is not above /foofoo/bar */
@@ -2596,28 +2842,35 @@ ctable_reinstall (CnxnTable* ct,
  */
 
 void          
-gconf_shutdown_daemon(GError** err)
+gconf_shutdown_daemon (GError** err)
 {
   CORBA_Environment ev;
   ConfigServer cs;
 
-  cs = gconf_get_config_server(FALSE, err); /* Don't want to spawn it if it's already down */
+  cs = gconf_get_config_server (FALSE, err); /* Don't want to spawn it if it's already down */
 
-  if (cs == CORBA_OBJECT_NIL)
+  if (err && *err && (*err)->code == GCONF_ERROR_NO_SERVER)
     {
-
+      /* No server is hardly an error here */
+      g_error_free (*err);
+      *err = NULL;
+    }
+  
+  if (cs == CORBA_OBJECT_NIL)
+    {      
+      
       return;
     }
 
-  CORBA_exception_init(&ev);
+  CORBA_exception_init (&ev);
 
-  ConfigServer_shutdown(cs, &ev);
+  ConfigServer_shutdown (cs, &ev);
 
   if (ev._major != CORBA_NO_EXCEPTION)
     {
       if (err)
-        *err = gconf_error_new(GCONF_ERROR_FAILED, _("Failure shutting down config server: %s"),
-                               CORBA_exception_id(&ev));
+        *err = gconf_error_new (GCONF_ERROR_FAILED, _("Failure shutting down config server: %s"),
+                                CORBA_exception_id (&ev));
 
       CORBA_exception_free(&ev);
     }
@@ -3173,26 +3426,26 @@ gconf_handle_corba_exception(CORBA_Environment* ev, GError** err)
   switch (ev->_major)
     {
     case CORBA_NO_EXCEPTION:
-      CORBA_exception_free(ev);
+      CORBA_exception_free (ev);
       return FALSE;
       break;
     case CORBA_SYSTEM_EXCEPTION:
       if (err)
-        *err = gconf_error_new(GCONF_ERROR_NO_SERVER, _("CORBA error: %s"),
-                               CORBA_exception_id(ev));
-      CORBA_exception_free(ev);
+        *err = gconf_error_new (GCONF_ERROR_NO_SERVER, _("CORBA error: %s"),
+                                CORBA_exception_id (ev));
+      CORBA_exception_free (ev);
       return TRUE;
       break;
     case CORBA_USER_EXCEPTION:
-      {
+      {        
         ConfigException* ce;
 
-        ce = CORBA_exception_value(ev);
+        ce = CORBA_exception_value (ev);
 
         if (err)
-          *err = gconf_error_new(corba_errno_to_gconf_errno(ce->err_no),
-                                 ce->message);
-        CORBA_exception_free(ev);
+          *err = gconf_error_new (corba_errno_to_gconf_errno (ce->err_no),
+                                  ce->message);
+        CORBA_exception_free (ev);
         return TRUE;
       }
       break;
